freeslot("MT_ANYWHO_GRAPHIC", "SPR_ANHG", "S_ANYWHO_SELECTED", "S_ANYWHO_ARROWS", "S_ANYWHO_GO", "S_ANYWHO_GOTEXT")
mobjinfo[MT_ANYWHO_GRAPHIC] = {
	spawnstate = S_INVISIBLE,
	radius = 1,
	height = 1,
	flags = MF_NOGRAVITY|MF_NOCLIPTHING
}

states[S_ANYWHO_SELECTED] = {
	sprite = SPR_ANHG,
	frame = A|FF_FLOORSPRITE,
	tics = -1
}

states[S_ANYWHO_ARROWS] = {
	sprite = SPR_ANHG,
	frame = B|FF_ANIMATE,
	var1 = B,
	var2 = TICRATE/3,
	tics = -1
}

states[S_ANYWHO_GO] = {
	sprite = SPR_ANHG,
	frame = D,
	tics = -1
}

states[S_ANYWHO_GOTEXT] = {
	sprite = SPR_ANHG,
	frame = E,
	tics = -1
}



local function foxBotTest()
	return not not CV_FindVar("ai_sys")
end

local function prettyFixed(fixed)
    local whole = fixed / FRACUNIT
    local part = abs(fixed % FRACUNIT)
    local str = tostring(whole) .. "." .. string.format("%04d", part*10000 / FRACUNIT)
    while string.sub(str, -1) == "0" do
        str = string.sub(str, 0, -2)
    end
    if string.sub(str, -1) == "." then
        str = string.sub(str, 0, -2) .. ".0"
    end
    return str
end

addHook("PostThinkFrame", function()
	if not multiplayer and displayplayer and displayplayer.mo and skins[displayplayer.skin].name == "anywho" and displayplayer.anywho then
		--lock camera
		local anywho = displayplayer.anywho
		if leveltime > 1 and anywho.camera then
			P_TeleportCameraMove(camera, unpack(anywho.camera.pos))
			camera.angle, camera.aiming = unpack(anywho.camera.angle)
		else
			anywho.camera = {
				pos = {
					camera.x,
					camera.y,
					camera.z
				},
				angle = {camera.angle, camera.aiming}
			}
		end
	end
	if multiplayer and leveltime > 0 then
		if server.mo and not server.anywhoMultiplayerNotification then
			server.anywhoMultiplayerNotification = true
			CONS_Printf(server, "\x85" + "Anywho shouldn't be loaded in multiplayer btw")
		end
	end
end)

addHook("PlayerCmd", function (p, cmd)
	if skins[p.skin].name ~= "anywho" then
		if p.anywhoHoldingJump then
			cmd.buttons = $ & ~BT_JUMP
		end
		return
	end
	if cmd.sidemove > 5 or cmd.buttons & BT_WEAPONNEXT then
		cmd.buttons = $ | BT_CUSTOM1
	end
	if cmd.sidemove < -5 or cmd.buttons & BT_WEAPONPREV then
		cmd.buttons = $ | BT_CUSTOM2
	end
	if cmd.buttons & BT_ATTACK then
		cmd.buttons = $ | BT_JUMP
	end
	if cmd.buttons & BT_FIRENORMAL then
		cmd.buttons = $ | BT_SPIN
	end
	cmd.buttons = $ & ~(BT_WEAPONPREV | BT_WEAPONNEXT)
end)

local FREEZEFLAGS = PF_FULLSTASIS | PF_GODMODE | PF_INVIS

addHook("PlayerThink",
--[[@param p player_t]]
function (p)
	if p.mo.skin ~= "anywho" then
		p.anywhoLastSkin = p.mo.skin
		if p.anywhoHoldingJump and not (p.cmd.buttons & BT_JUMP) then
			p.anywhoHoldingJump = false
		end
		return
	end
	
	if multiplayer then
		local prevSkin = p.anywhoLastSkin or "sonic"
		R_SetPlayerSkin(p, prevSkin)
		p.mo.skin = prevSkin
		CONS_Printf(p, "You shouldn't be selecting Anywho in multiplayer.")
		return
	end
	
	if not players[1] then
		if modeattacking then
			if displayplayer.playerstate == PST_LIVE then
				CONS_Printf(p, "\x85" + "Anywho doesn't yet work for Record Attack/NiGHTS runs.")
				P_KillMobj(displayplayer.mo)
				displayplayer.deadtimer = TICRATE*2
			end
			return
		end
	end
	
	if not consoleplayer.anywho and leveltime > 0 then
		CONS_Printf(p, "Using Anywho in any way besides the character select is janky.")
		CONS_Printf(p, "Colors will be wack until the save is reloaded and your P2 selection usually won't stick.")
		CONS_Printf(p, "I'll let you do it,\x87\x8f but don't say I didn't warn you.")
	end
	
	p.pflags = $ | FREEZEFLAGS
	
	if not consoleplayer.anywho then
		hud.disable("score")
		hud.disable("time")
		hud.disable("rings")
		hud.disable("lives")
		
		local validSkinCount = 0
		while validSkinCount < 31 and skins[validSkinCount+1] do
			validSkinCount = $ + 1
		end
		consoleplayer.anywho = {
			pos = {p.mo.x, p.mo.y, p.mo.z},
			angle = p.mo.angle,
			skin1 = 0,
			skin2 = 1,
			selectState = -1,
			skinCount = validSkinCount,
			mobjs = {
				selected = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_ANYWHO_GRAPHIC),
				arrows = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_ANYWHO_GRAPHIC),
				go = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_ANYWHO_GRAPHIC),
				gotext = P_SpawnMobj(0, 0, 0, MT_OVERLAY)
			}
		}
		local aw = consoleplayer.anywho
		
		aw.mobjs.selected.renderflags = $ | RF_SHADOWDRAW
		
		aw.mobjs.go.color = SKINCOLOR_BLUE
		
		aw.mobjs.gotext.target = aw.mobjs.go
		aw.mobjs.gotext.state = S_INVISIBLE
		aw.mobjs.gotext.color = SKINCOLOR_ORANGE
	end
	local anywho = consoleplayer.anywho
	
	if players[1].mo.skin ~= "anywho" then
		R_SetPlayerSkin(players[1], "anywho")
		players[1].mo.skin = "anywho"
	end
	
	if leveltime == 3 then
		S_StartSound(p.mo, sfx_appear, p)
	end
	
	p.mo.flags2 = $ | MF2_DONTDRAW
	
	local justMadePlayer = false
	if not p.anywhoPlayer then
		justMadePlayer = true
		local overlay = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_OVERLAY)
		overlay.target = p.mo
		overlay.state = S_PLAY_STND
		overlay.skin = "sonic"
		overlay.color = SKINCOLOR_BLUE
		p.anywhoPlayer = {
			overlay = overlay
		}
	end
	local awPlayer = p.anywhoPlayer
	
	if leveltime <= 0 or justMadePlayer then
		awPlayer.overlay.flags2 = $ | MF2_DONTDRAW
		awPlayer.overlay.state = S_PLAY_STND
		awPlayer.overlay.frame = ($ | FF_TRANS90)
		awPlayer.transparency = 9
		anywho.selectState = -1
		P_SetOrigin(p.mo, unpack(anywho.pos))
		return
	else
		if p == consoleplayer then
			p.drawangle = anywho.angle + ANGLE_180 -- + ANGLE_45
		else
			p.drawangle = anywho.angle + ANGLE_180 -- - ANGLE_45
		end
		if leveltime > 3 and leveltime % 2 == 0 then
			awPlayer.overlay.flags2 = $ & ~MF2_DONTDRAW
			awPlayer.transparency = max($ - 1, 0)
		end
		
		local transbit = FF_TRANS10 * awPlayer.transparency
		awPlayer.overlay.frame = ($ & ~FF_TRANSMASK) | transbit
		if leveltime > TICRATE/2 and anywho.selectState < 0 then
			anywho.selectState = 0
		end
	end
	
	local chosenSkin = nil --skins[P_RandomKey(validSkinCount)].name
	
	if anywho.selectState < 0 then
		anywho.mobjs.selected.state = S_INVISIBLE
		anywho.mobjs.arrows.state = S_INVISIBLE
		anywho.mobjs.go.state = S_INVISIBLE
		anywho.mobjs.gotext.state = S_INVISIBLE
		anywho.mobjs.go.frame = $ | FF_TRANS50
		anywho.mobjs.gotext.frame = $ | FF_TRANS50
	else
		anywho.mobjs.selected.state = S_ANYWHO_SELECTED
		anywho.mobjs.go.state = S_ANYWHO_GO
		anywho.mobjs.gotext.state = S_ANYWHO_GOTEXT
		if anywho.selectState < 2 then
			if anywho.mobjs.arrows.state ~= S_ANYWHO_ARROWS then
				anywho.mobjs.arrows.state = S_ANYWHO_ARROWS
			end
			anywho.mobjs.go.color = SKINCOLOR_GREY
			anywho.mobjs.gotext.color = SKINCOLOR_GREY
			anywho.mobjs.go.frame = $ | FF_TRANS50
			anywho.mobjs.gotext.frame = $ | FF_TRANS50
		else
			anywho.mobjs.arrows.state = S_INVISIBLE
			anywho.mobjs.go.frame = $ & ~FF_TRANS50
			anywho.mobjs.gotext.frame = $ & ~FF_TRANS50
			anywho.mobjs.go.color = skins[anywho.skin1].prefcolor
			anywho.mobjs.gotext.color = skins[anywho.skin2].prefcolor
			--anywho.mobjs.go.color = skincolors[skins[anywho.skin1].prefcolor].invcolor
			--anywho.mobjs.gotext.color = skins[anywho.skin1].prefcolor
			
			
			
		end
	end
	
	local targetPos, mySkinNum
	if p == consoleplayer then
		--controls
		local cmd = p.lastCmd or p.cmd
		
		if p.prevcmd and anywho.selectState >= 0 then
			if (cmd.buttons & BT_JUMP) and not (p.prevcmd.buttons & BT_JUMP) then
				local sounds = {sfx_hoop1, sfx_hoop2, sfx_ncitem} --sfx_hoop3}
				anywho.selectState = $ + 1
				S_StartSound(p.mo, sounds[anywho.selectState], p)
			end
			if (cmd.buttons & BT_SPIN) and not (p.prevcmd.buttons & BT_SPIN) then
				if anywho.selectState > 0 then
					S_StartSound(p.mo, sfx_iceb, p) --sfx_ncchip
					anywho.selectState = $ - 1
				end
			end
			local wepmask = cmd.buttons & (BT_CUSTOM1 | BT_CUSTOM2)
			if anywho.selectState < 2
			  and (p.prevcmd.buttons & (BT_CUSTOM1 | BT_CUSTOM2)) == 0
			  and (wepmask == BT_CUSTOM1 or wepmask == BT_CUSTOM2) then
				local currentSkin = ({anywho.skin1, anywho.skin2})[anywho.selectState + 1]
				S_StartSound(p.mo, sfx_s3kb7, p)
				if wepmask == BT_CUSTOM1 then
					currentSkin = $ + 1
					if currentSkin >= 0 and currentSkin < 32 and skins[currentSkin] and skins[currentSkin].name == "anywho" then
						currentSkin = $ + 1
					end
				else
					currentSkin = $ - 1
					if currentSkin >= 0 and currentSkin < 32 and skins[currentSkin] and skins[currentSkin].name == "anywho" then
						currentSkin = $ - 1
					end
				end
				if currentSkin < 0 then
					currentSkin = anywho.skinCount
				elseif currentSkin > anywho.skinCount then
					currentSkin = 0
				end
				if skins[currentSkin].name == "anywho" then
					if wepmask == BT_CUSTOM1 then
						currentSkin = $ + 1
						
					else
						currentSkin = $ - 1
					end
				end
				if anywho.selectState == 0 then
					anywho.skin1 = currentSkin
				else
					anywho.skin2 = currentSkin
				end
			end
		end
		
		p.prevcmd = {
			buttons = cmd.buttons,
			side = cmd.sidemove
		}
		--controls end
		targetPos = {
			anywho.pos[1] - (sin(anywho.angle)*50),
			anywho.pos[2] + (cos(anywho.angle)*50),
			anywho.pos[3]
		}
		if anywho.selectState == 0 then
			P_SetOrigin(
				anywho.mobjs.selected,
				p.mo.x, p.mo.y, p.mo.floorz
			)
			P_SetOrigin(
				anywho.mobjs.arrows,
				p.mo.x, p.mo.y, p.mo.floorz
			)
			if awPlayer.overlay.state ~= S_PLAY_WALK then
				awPlayer.overlay.state = S_PLAY_WALK
			end
		else
			if awPlayer.overlay.state ~= S_PLAY_STND then
				awPlayer.overlay.state = S_PLAY_STND
			end
			awPlayer.overlay.tics = TICRATE
		end
		mySkinNum = anywho.skin1
	else
		targetPos = {
			anywho.pos[1] - (sin(anywho.angle)*-50),
			anywho.pos[2] + (cos(anywho.angle)*-50),
			anywho.pos[3]
		}
		if anywho.selectState == 1 then
			P_SetOrigin(
				anywho.mobjs.selected,
				p.mo.x, p.mo.y, p.mo.floorz
			)
			P_SetOrigin(
				anywho.mobjs.arrows,
				p.mo.x, p.mo.y, p.mo.floorz
			)
			if awPlayer.overlay.state ~= S_PLAY_WALK then
				awPlayer.overlay.state = S_PLAY_WALK
			end
		else
			if awPlayer.overlay.state ~= S_PLAY_STND then
				awPlayer.overlay.state = S_PLAY_STND
			end
			awPlayer.overlay.tics = TICRATE
		end
		mySkinNum = anywho.skin2
	end
	if anywho.selectState >= 3 then
		chosenSkin = skins[mySkinNum].name
	end
	if awPlayer.lastSkinNum ~= mySkinNum then
		awPlayer.overlay.skin = skins[mySkinNum].name
		awPlayer.overlay.color = skins[mySkinNum].prefcolor
		awPlayer.overlay.state = S_PLAY_WALK
		awPlayer.lastSkinNum = mySkinNum
	end
	if anywho.selectState >= 2 then
		P_SetOrigin(
			anywho.mobjs.selected,
			unpack(anywho.pos)
		)
	end
	P_MoveOrigin(
		p.mo,
		p.mo.x + (targetPos[1] - p.mo.x)/4,
		p.mo.y + (targetPos[2] - p.mo.y)/4,
		p.mo.z
	)
	
	if chosenSkin then
		R_SetPlayerSkin(p, chosenSkin)
		p.mo.skin = chosenSkin
		p.pflags = $ & ~FREEZEFLAGS
		hud.enable("score")
		hud.enable("time")
		hud.enable("rings")
		hud.enable("lives")
		P_RemoveMobj(awPlayer.overlay)
		p.anywhoHoldingJump = true
		if p ~= consoleplayer then -- final touches
			P_RemoveMobj(anywho.mobjs.selected)
			P_RemoveMobj(anywho.mobjs.arrows)
			P_RemoveMobj(anywho.mobjs.go)
		end
	end
	
end)

addHook("JumpSpecial", function (p)
	if p.anywhoHoldingJump then
		return true
	end
end)

local centerx = 160*FRACUNIT
local centery = 100*FRACUNIT

--[[hud.add(function (v, p, cam)
	-- i cant detect the fov nor figure out how to align the nametag given a fixed FOV
	if p.mo.skin == "anywho" and p.anywho and p.anywho.selectState >= 0 then
		local dupx = v.width() * FRACUNIT / 320
		local dupy = v.height() * FRACUNIT / 200
		local scalefac = 5*(max(FixedDiv(dupx, dupy), FixedDiv(dupy, dupx))-FRACUNIT)/2 + FRACUNIT
		local scalefac = max(FixedDiv(dupx, dupy), FixedDiv(dupy, dupx))
		scalefac = FixedMul(scalefac, scalefac)
		scalefac = FRACUNIT
		print(prettyFixed(dupx), prettyFixed(dupy), prettyFixed(FixedDiv(dupx, dupy)))
		local offx = FixedMul((41)*FRACUNIT, scalefac)
		local offy = FixedMul((59)*FRACUNIT, scalefac)
		v.drawString(
			centerx-offx, centery+offy,
			"oio", --skins[p.anywho.skin1].realname,
			skincolors[skins[p.anywho.skin1].prefcolor].chatcolor|V_SNAPTOLEFT|V_SNAPTOBOTTOM|V_ALLOWLOWERCASE,
			"thin-fixed-center"
		)
		v.drawString(
			centerx+(41*FRACUNIT), centery+(59*FRACUNIT),
			"oio", --skins[p.anywho.skin2].realname,
			skincolors[skins[p.anywho.skin2].prefcolor].chatcolor|V_SNAPTORIGHT|V_ALLOWLOWERCASE,
			"thin-fixed-center"
		)
	end
end, "game")]]

addHook("GameQuit", function ()
	if consoleplayer.mo and consoleplayer.mo.skin == "anywho" then
		hud.enable("score")
		hud.enable("time")
		hud.enable("rings")
		hud.enable("lives")
	end
end)